6  # -*- coding: utf-8 -*-
"""
Created on Mon Jun 17 11:19:37 2013

@author: proto
"""

import libsbml
import json
from optparse import OptionParser


def factorial(x):
    temp = x
    acc = 1
    while temp > 0:
        acc *= temp
        temp -= 1
    return acc


def comb(x, y, exact=True):
    return factorial(x) / (factorial(y) * factorial(x - y))


class SBML2JSON:
    def __init__(self, model):
        self.model = model
        self.getUnits()
        self.moleculeData = {}

    def getUnits(self):
        self.unitDictionary = {}
        # unitDictionary['substance'] = [libsbml.UNIT_KIND_MOLE,1]
        # unitDictionary['volume'] = [libsbml.UNIT_KIND_LITER,1]
        # unitDictionary['area'] = [libsbml.UNIT_KIND_METER,1]
        # unitDictionary['length'] = [libsbml.UNIT_KIND_METER,1]
        # unitDictionary['time'] = [libsbml.UNIT_KIND_SECOND,1]

        for unitDefinition in self.model.getListOfUnitDefinitions():
            unitList = []
            for unit in unitDefinition.getListOfUnits():

                unitList.append([unit.getKind(), unit.getScale(), unit.getExponent()])

            self.unitDictionary[unitDefinition.getId()] = unitList
        """
        
Table 3: SBML's built-in units.
Name	 Possible Scalable Units	 Default Units	
substance	 mole, item	 mole	
volume	 litre, cubic metre	 litre	
area	 square metre	 square metre	
length	 metre	 metre	
time	 second	 second	
"""

    def getParameters(self):
        parameters = []
        prx = {
            "name": "Nav",
            "value": "6.022e8",
            "unit": "",
            "type": "Avogadro number for 1 um^3",
        }
        parameters.append(prx)
        for parameter in self.model.getListOfParameters():
            parameterSpecs = {
                "name": parameter.getId(),
                "value": parameter.getValue(),
                "unit": parameter.getUnits(),
                "type": "",
            }
            """                             
            if parameterSpecs[0] == 'e':
                parameterSpecs = ('are',parameterSpecs[1])
            if parameterSpecs[1] == 0:
                zparam.append(parameterSpecs[0])
            """
            if parameter.getUnits() in self.unitDictionary:
                for factor in self.unitDictionary[parameter.getUnits()]:
                    parameterSpecs["value"] *= 10 ** (factor[1] * factor[2])
                    parameterSpecs["unit"] = "{0}*1e{1}".format(
                        parameterSpecs["unit"], factor[1] * factor[2]
                    )
                if (
                    "mole" in parameter.getUnits()
                    and "per_mole" not in parameter.getUnits()
                ):
                    parameterSpecs["value"] *= float(6.022e8)
                    parameterSpecs["unit"] = "{0}*{1}".format(
                        parameterSpecs["unit"], "avo.num"
                    )
            # if parameter.getUnits() == '':
            #    parameterSpecs['value'] *= float(6.022e8*1000)
            #    parameterSpecs['unit'] = '{0}*{1}'.format(parameterSpecs['unit'],'avo.num*1000')

            parameters.append(parameterSpecs)
        prx = {"name": "rxn_layer_t", "value": "0.01", "unit": "um", "type": ""}
        ph = {"name": "h", "value": "rxn_layer_t", "unit": "um", "type": ""}
        pRs = {"name": "Rs", "value": "0.002564", "unit": "um", "type": ""}
        pRc = {"name": "Rc", "value": "0.0015", "unit": "um", "type": ""}
        parameters.append(prx)
        parameters.append(ph)
        parameters.append(pRs)
        parameters.append(pRc)
        parameterDict = {idx + 1: x for idx, x in enumerate(parameters)}
        return parameterDict

    def __getRawCompartments(self):
        """
        extracts information about the compartments in a model
        *returns* name,dimensions,size
        """
        compartmentList = {}
        for compartment in self.model.getListOfCompartments():
            name = compartment.getId()
            size = compartment.getSize()
            outside = compartment.getOutside()
            dimensions = compartment.getSpatialDimensions()
            compartmentList[name] = [dimensions, size, outside]
        return compartmentList

    def getOutsideInsideCompartment(self, compartmentList, compartment):
        """
        Gets the containing compartment for this compartment
        """
        outside = compartmentList[compartment][2]
        for comp in compartmentList:
            if compartmentList[comp][2] == compartment:
                return outside, comp
        return outside, -1

    def getMolecules(self):
        """
        *species* is the element whose SBML information we will extract
        this method gets information directly
        from an SBML related to a particular species.
        It returns id,initialConcentration,(bool)isconstant and isboundary,
        and the compartment
        """

        compartmentList = self.__getRawCompartments()

        molecules = []
        release = []
        for idx, species in enumerate(self.model.getListOfSpecies()):
            compartment = species.getCompartment()
            if compartmentList[compartment][0] == 3:
                typeD = "3D"
                outside, inside = self.getOutsideInsideCompartment(
                    compartmentList, compartment
                )
                diffusion = "KB*T/(6*PI*mu_{0}*Rs)".format(compartment)
            else:
                typeD = "2D"

                diffusion = "KB*T*LOG((mu_{0}*h/(SQRT(4)*Rc*(mu_{1}+mu_{2})/2))-gamma)/(4*PI*mu_{0}*h)".format(
                    compartment, outside, inside
                )
            self.moleculeData[species.getId()] = [compartmentList[compartment][0]]
            moleculeSpecs = {
                "name": species.getId(),
                "type": typeD,
                "extendedName": species.getName(),
                "dif": diffusion,
            }
            initialConcentration = species.getInitialConcentration()

            if initialConcentration == 0:
                initialConcentration = species.getInitialAmount()
            if species.getSubstanceUnits() in self.unitDictionary:
                for factor in self.unitDictionary[species.getSubstanceUnits()]:
                    initialConcentration *= 10 ** (factor[1] * factor[2])
                if "mole" in species.getSubstanceUnits():
                    initialConcentration /= float(6.022e8)
            if species.getSubstanceUnits() == "":
                initialConcentration /= float(6.022e8)

            isConstant = species.getConstant()
            # isBoundary = species.getBoundaryCondition()
            if initialConcentration != 0:
                if compartmentList[compartment][0] == 2:
                    objectExpr = "{0}[{1}]".format(inside.upper(), compartment.upper())
                else:
                    objectExpr = "{0}".format(compartment)
                releaseSpecs = {
                    "name": "Release_Site_s{0}".format(idx + 1),
                    "molecule": species.getId(),
                    "shape": "OBJECT",
                    "quantity_type": "NUMBER_TO_RELEASE",
                    "quantity_expr": initialConcentration,
                    "object_expr": objectExpr,
                }
                release.append(releaseSpecs)
            # self.speciesDictionary[identifier] = standardizeName(name)
            # returnID = identifier if self.useID else \
            molecules.append(moleculeSpecs)

            # self.sp eciesDictionary[identifier]
        moleculesDict = {idx + 1: x for idx, x in enumerate(molecules)}
        releaseDict = {idx + 1: x for idx, x in enumerate(release)}
        return moleculesDict, releaseDict

    def getPrunnedTree(self, math, remainderPatterns):
        """
        remove mass action factors
        """
        while (math.getCharacter() == "*" or math.getCharacter() == "/") and len(
            remainderPatterns
        ) > 0:
            if libsbml.formulaToString(math.getLeftChild()) in remainderPatterns:
                remainderPatterns.remove(libsbml.formulaToString(math.getLeftChild()))
                math = math.getRightChild()
            elif libsbml.formulaToString(math.getRightChild()) in remainderPatterns:
                remainderPatterns.remove(libsbml.formulaToString(math.getRightChild()))
                math = math.getLeftChild()
            else:
                if (math.getLeftChild().getCharacter()) == "*":
                    math.replaceChild(
                        0, self.getPrunnedTree(math.getLeftChild(), remainderPatterns)
                    )
                if (math.getRightChild().getCharacter()) == "*":
                    math.replaceChild(
                        math.getNumChildren() - 1,
                        self.getPrunnedTree(math.getRightChild(), remainderPatterns),
                    )
                break
        return math

    def getInstanceRate(self, math, compartmentList, reversible, rReactant, rProduct):

        # remove compartments from expression
        math = self.getPrunnedTree(math, compartmentList)

        if reversible:
            if math.getCharacter() == "-" and math.getNumChildren() > 1:
                rateL, nl = self.removeFactorFromMath(
                    math.getLeftChild().deepCopy(), rReactant, rProduct
                )
                rateR, nr = self.removeFactorFromMath(
                    math.getRightChild().deepCopy(), rProduct, rReactant
                )
            else:
                rateL, nl = self.removeFactorFromMath(math, rReactant, rProduct)
                rateL = "if({0} >= 0 ,{0},0)".format(rateL)
                rateR, nr = self.removeFactorFromMath(math, rReactant, rProduct)
                rateR = "if({0} < 0 ,-({0}),0)".format(rateR)
                nl, nr = 1, 1
        else:
            rateL, nl = self.removeFactorFromMath(math.deepCopy(), rReactant, rProduct)
            rateR, nr = "0", "-1"
        if reversible:
            pass
        return rateL, rateR

    def removeFactorFromMath(self, math, reactants, products):

        remainderPatterns = []
        highStoichoiMetryFactor = 1
        for x in reactants:
            highStoichoiMetryFactor *= factorial(x[1])
            y = [i[1] for i in products if i[0] == x[0]]
            y = y[0] if len(y) > 0 else 0
            # TODO: check if this actually keeps the correct dynamics
            # this is basically there to address the case where theres more products
            # than reactants (synthesis)
            if x[1] > y:
                highStoichoiMetryFactor /= comb(int(x[1]), int(y), exact=True)
            for counter in range(0, int(x[1])):
                remainderPatterns.append(x[0])
        # for x in products:
        #    highStoichoiMetryFactor /= math.factorial(x[1])
        # remainderPatterns = [x[0] for x in reactants]
        math = self.getPrunnedTree(math, remainderPatterns)
        rateR = libsbml.formulaToString(math)
        for element in remainderPatterns:
            rateR = "if({0} >0,({1})/{0} ,0)".format(element, rateR)
        if highStoichoiMetryFactor != 1:
            rateR = "{0}*{1}".format(rateR, int(highStoichoiMetryFactor))

        return rateR, math.getNumChildren()

    def adjustParameters(self, stoichoimetry, rate, parameters):
        """
        adds avogadros number and other adjusting factors to the reaction rates
        """
        for parameter in parameters:
            if (
                parameters[parameter]["name"] in rate
                and parameters[parameter]["unit"] == ""
            ):
                print(parameters[parameter])
                if stoichoimetry == 2:
                    parameters[parameter]["value"] *= float(6.022e8)
                    parameters[parameter]["unit"] = "Bimolecular * NaV"
                elif stoichoimetry == 0:
                    parameters[parameter]["value"] /= float(6.022e8)
                    parameters[parameter]["unit"] = "0-order / NaV"
                elif stoichoimetry == 1:
                    parameters[parameter]["unit"] = "Unimolecular"
                print(parameters[parameter])

    def getReactions(self, sparameters):
        """
        returns a list with reactant,product and fwdRate
        """

        reactionSpecs = []
        for index, reaction in enumerate(self.model.getListOfReactions()):
            reactant = [
                (reactant.getSpecies(), reactant.getStoichiometry())
                for reactant in reaction.getListOfReactants()
                if reactant.getSpecies() != "EmptySet"
            ]
            product = [
                (product.getSpecies(), product.getStoichiometry())
                for product in reaction.getListOfProducts()
                if product.getSpecies() != "EmptySet"
            ]

            kineticLaw = reaction.getKineticLaw()
            rReactant = [
                (x.getSpecies(), x.getStoichiometry())
                for x in reaction.getListOfReactants()
                if x.getSpecies() != "EmptySet"
            ]
            rProduct = [
                (x.getSpecies(), x.getStoichiometry())
                for x in reaction.getListOfProducts()
                if x.getSpecies() != "EmptySet"
            ]

            parameters = [
                (parameter.getId(), parameter.getValue())
                for parameter in kineticLaw.getListOfParameters()
            ]

            math = kineticLaw.getMath()
            reversible = reaction.getReversible()
            compartmentList = []
            for compartment in self.model.getListOfCompartments():
                compartmentList.append(compartment.getId())

            rateL, rateR = self.getInstanceRate(
                math, compartmentList, reversible, rReactant, rProduct
            )
            # finalReactant = [x[0]]
            # testing whether we have volume-surface interactions
            rcList = []
            prdList = []
            for element in reactant:
                orientation = (
                    ","
                    if len(set(self.moleculeData[x[0]][0] for x in reactant)) > 1
                    and self.moleculeData[element[0]] == "3"
                    else "'"
                )
                rcList.append("{0}{1}".format(element[0], orientation))
            for element in product:
                orientation = (
                    ","
                    if len(set(self.moleculeData[x[0]][0] for x in reactant)) > 1
                    and self.moleculeData[element[0]] == "3"
                    else "'"
                )

                prdList.append("{0}{1}".format(element[0], orientation))
            if rateL != 0:
                tmp = {}
                tmp["reactants"] = " + ".join(rcList)
                tmp["products"] = " + ".join(prdList)
                tmp["fwd_rate"] = rateL
                reactionSpecs.append(tmp)
            if rateR != 0:
                tmp = {}
                tmp["reactants"] = " + ".join(prdList)
                tmp["products"] = " + ".join(rcList)
                tmp["fwd_rate"] = rateR
                reactionSpecs.append(tmp)

            self.adjustParameters(len(reactant), rateL, sparameters)
            self.adjustParameters(len(product), rateR, sparameters)
        reactionDict = {idx + 1: x for idx, x in enumerate(reactionSpecs)}
        return reactionDict
        # SBML USE INSTANCE RATE
        # HOW TO GET THE DIFFUSION CONSTANT


def main():

    # command line arguments
    parser = OptionParser()
    parser.add_option(
        "-i",
        "--input",
        dest="input",
        default="bngl2mcell/rec_dim_sbml.xml",
        type="string",
        help="The input SBML file in xml format. Default = 'input.xml'",
        metavar="FILE",
    )
    parser.add_option(
        "-o",
        "--output",
        dest="output",
        type="string",
        help="the output JSON file. Default = <input>.py",
        metavar="FILE",
    )
    (options, args) = parser.parse_args()
    reader = libsbml.SBMLReader()
    nameStr = options.input
    if options.output == None:
        outputFile = nameStr + ".py"
    else:
        outputFile = options.output

    # libsbml initialization stuff
    document = reader.readSBMLFromFile(nameStr)
    if document.getModel() == None:
        print("No such input file")
        return
    # get data
    parser = SBML2JSON(document.getModel())
    parameters = parser.getParameters()
    molecules, release = parser.getMolecules()
    reactions = parser.getReactions(parameters)
    definition = {}
    definition["par_list"] = parameters
    definition["mol_list"] = molecules
    definition["rxn_list"] = reactions
    definition["rel_list"] = release
    print("Writing output to {0}".format(outputFile))
    # output
    with open(outputFile, "w") as f:
        json.dump(definition, f, sort_keys=True, indent=1, separators=(",", ": "))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Created on Tue May 21 12:38:21 2013

@author: proto
"""

import libsbml2bngl as ls2b
import argparse
import yaml


def defineConsole():
    parser = argparse.ArgumentParser(description="SBML to BNGL translator")
    parser.add_argument(
        "-i", "--input-file", type=str, help="input SBML file", required=True
    )
    parser.add_argument(
        "-t", "--annotation", action="store_true", help="keep annotation information"
    )
    parser.add_argument("-o", "--output-file", type=str, help="output SBML file")
    parser.add_argument("-c", "--convention-file", type=str, help="Conventions file")
    parser.add_argument(
        "-n", "--naming-conventions", type=str, help="Naming conventions file"
    )
    parser.add_argument(
        "-u", "--user-structures", type=str, help="User defined species"
    )
    parser.add_argument(
        "-id",
        "--molecule-id",
        action="store_true",
        help="use SBML molecule ids instead of names. IDs are less descriptive but more bngl friendly. Use only if the generated BNGL has syntactic errors",
    )
    parser.add_argument(
        "-cu",
        "--convert-units",
        action="store_false",
        help="convert units. Otherwise units are copied straight from sbml to bngl",
    )
    parser.add_argument(
        "-a", "--atomize", action="store_true", help="Infer molecular structure"
    )
    parser.add_argument(
        "-p",
        "--pathwaycommons",
        action="store_true",
        help="Use pathway commons to infer molecule binding. This setting requires an internet connection and will query the pathway commons web service.",
    )
    parser.add_argument(
        "-b",
        "--bionetgen-analysis",
        type=str,
        help="Set the BioNetGen path for context post analysis.",
    )
    parser.add_argument(
        "-s",
        "--isomorphism-check",
        action="store_true",
        help="disallow atomizations that produce the same graph structure",
    )
    parser.add_argument(
        "-I",
        "--ignore",
        action="store_true",
        help="ignore atomization translation errors",
    )
    parser.add_argument(
        "-mr",
        "--memoized-resolver",
        default=False,
        action="store_true",
        help="sometimes the dependency graph is too large and might cause a very large memory requirement. This option will slow the translator down but will decrease memory usage",
    )
    parser.add_argument(
        "-k",
        "--keep-local-parameters",
        default=False,
        action="store_true",
        help="this option will keep the local parameters unresolved so that they can be controlled from the parameter section in the BNGL. Without this option, local parameters will be resolved to their values in functions",
    )
    parser.add_argument(
        "-q",
        "--quiet-mode",
        default=False,
        action="store_true",
        help="this option will supress logging into STDIO and instead will write the logging into a file",
    )
    parser.add_argument(
        "-ll",
        "--log-level",
        default="DEBUG",
        help='This option allows you to select a logging level, from quietest to loudest options are: "CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG". Default is set to DEBUG',
    )

    return parser


def checkInput(namespace):
    options = {}
    options["inputFile"] = namespace.input_file

    conv, useID, naming = ls2b.selectReactionDefinitions(options["inputFile"])
    options["outputFile"] = (
        namespace.output_file
        if namespace.output_file is not None
        else options["inputFile"] + ".bngl"
    )
    options["conventionFile"] = (
        namespace.convention_file if namespace.convention_file is not None else conv
    )
    options["userStructure"] = namespace.user_structures
    options["namingConventions"] = (
        namespace.naming_conventions
        if namespace.naming_conventions is not None
        else naming
    )
    options["useId"] = namespace.molecule_id
    options["annotation"] = namespace.annotation
    options["atomize"] = namespace.atomize
    options["pathwaycommons"] = namespace.pathwaycommons
    options["bionetgenAnalysis"] = namespace.bionetgen_analysis
    options["isomorphismCheck"] = namespace.isomorphism_check
    options["ignore"] = namespace.ignore
    options["noConversion"] = namespace.convert_units
    options["memoizedResolver"] = namespace.memoized_resolver
    options["replaceLocParams"] = not namespace.keep_local_parameters
    options["quietMode"] = namespace.quiet_mode
    assert namespace.log_level in [
        "CRITICAL",
        "ERROR",
        "WARNING",
        "INFO",
        "DEBUG",
    ], "Logging level {} is not an allowed level".format(namespace.log_level)
    options["logLevel"] = namespace.log_level
    return options


def main():
    parser = defineConsole()
    namespace = parser.parse_args()

    options = checkInput(namespace)
    returnArray = ls2b.analyzeFile(
        options["inputFile"],
        options["conventionFile"],
        options["useId"],
        options["namingConventions"],
        options["outputFile"],
        speciesEquivalence=options["userStructure"],
        atomize=options["atomize"],
        bioGrid=False,
        pathwaycommons=options["pathwaycommons"],
        ignore=options["ignore"],
        noConversion=options["noConversion"],
        memoizedResolver=options["memoizedResolver"],
        replaceLocParams=options["replaceLocParams"],
        quietMode=options["quietMode"],
        logLevel=options["logLevel"],
    )

    if namespace.bionetgen_analysis and returnArray:
        ls2b.postAnalyzeFile(
            options["outputFile"],
            namespace.bionetgen_analysis,
            returnArray.database,
            replaceLocParams=options["replaceLocParams"],
        )

    if namespace.annotation and returnArray:
        with open(options["outputFile"] + ".yml", "w") as f:
            f.write(yaml.dump(returnArray.annotation, default_flow_style=False))


if __name__ == "__main__":
    main()



try:
    from typing import OrderedDict
except ImportError:
    from collections import OrderedDict
from .structs import Parameter, Compartment, Observable
from .structs import MoleculeType, Species, Function
from .structs import Rule, Action
from .utils import ActionList

# this import fails on some python versions
try:
    from typing import OrderedDict
except ImportError:
    from collections import OrderedDict

###### BLOCK OBJECTS ######
class ModelBlock:
    """
    Base block object that will be used for each block in the model.

    Attributes
    ----------
    name : str
        Name of the block which will be used to write the BNGL text
    comment : (str, str)
        comment at the begin {block} or end {block} statements, tuple
    items : OrderedDict
        all the model objects in the block
    _changes : OrderedDict
        a dictionary to keep track of all the changes done in a block
        after it is originally created
    _recompile : bool
        a tag that tells a potential future simulator if the model
        needs to be recompiled. TODO: has to be computed from _changes
        property upon get request.

    Methods
    -------
    reset_compilation_tags()
        resets _recompile and _changes tags for the block
    add_item((name,value))
        sets self.item[name] = value to add a particular model object
        into a block
    add_items(item_list)
        loops over every element in the list and uses add_item on it
    gen_string()
        for every block this method generates the BNGL string of the
        block. it has to be overwritten for each block.
    """

    def __init__(self) -> None:
        self.name = "ModelBlock"
        self.comment = (None, None)
        self._changes = OrderedDict()
        self._recompile = False
        self.items = OrderedDict()

    def __str__(self) -> str:
        return self.gen_string()

    def __len__(self) -> int:
        return len(self.items)

    def __repr__(self) -> str:
        # overwrites what the class representation
        # shows the items in the model block in
        # say ipython
        repr_str = "{} block with {} item(s): {}".format(
            self.name, len(self.items), list([i.name for i in self.items.values()])
        )
        return repr_str

    def __getitem__(self, key):
        if isinstance(key, int):
            # get the item in order
            return list(self.items.keys())[key]
        return self.items[key]

    def __setitem__(self, key, value) -> None:
        self.items[key] = value

    def __delitem__(self, key) -> None:
        if key in self.items:
            self.items.pop(key)
        else:
            print("Item {} not found".format(key))

    def __iter__(self):
        return self.items.keys().__iter__()

    def __contains__(self, key) -> bool:
        return key in self.items

    # TODO: Think extensively how this is going to work
    def __setattr__(self, name, value) -> None:
        changed = False
        if hasattr(self, "items"):
            if name in self.items.keys():
                try:
                    new_value = float(value)
                    changed = True
                    self.items[name] = new_value
                except:
                    self.items[name] = value
                if changed:
                    self._changes[name] = new_value
                    self.__dict__[name] = new_value
        else:
            self.__dict__[name] = value

    def gen_string(self) -> str:
        # each block can have a comment at the start
        if self.comment[0] is not None:
            block_lines = ["\nbegin {} #{}".format(self.name, self.comment[0])]
        else:
            block_lines = ["\nbegin {}".format(self.name)]
        # now we just loop over lines
        for item in self.items.keys():
            block_lines.append(self.items[item].print_line())
        # each block can have a comment at the start
        if self.comment[1] is not None:
            block_lines.append("end {} #{}\n".format(self.name, self.comment[1]))
        else:
            block_lines.append("end {}\n".format(self.name))
        # join everything with new lines
        return "\n".join(block_lines)

    def reset_compilation_tags(self) -> None:
        # TODO: Make these properties such that it checks each
        # item for changes/recompile tags
        # for item in self.items:
        #     self.items[item]._recompile = False
        #     self.items[item]._changes = {}
        self._changes = OrderedDict()
        self._recompile = False

    def add_item(self, item_tpl) -> None:
        # TODO: try adding evaluation of the parameter here
        # for the future, in case we want people to be able
        # to adjust the math
        # TODO: Error handling, some names will definitely break this
        name, value = item_tpl
        # allow for empty addition, uses index
        if name is None:
            name = len(self.items)
        # set the line
        self.items[name] = value
        # if the name is a string, try adding as an attribute
        if isinstance(name, str):
            try:
                setattr(self, name, value)
            except:
                print("can't set {} to {}".format(name, value))
                pass
        # we just added an item to a block, let's assume we need
        # to recompile if we have a compiled simulator
        self._recompile = True

    def add_items(self, item_list) -> None:
        for item in item_list:
            self.add_item(item)


class ParameterBlock(ModelBlock):
    """
    Parameter block object, subclass of ModelBlock.

    Methods
    -------
    add_parameter(name, value, expr=None)
        adds a parameter by making a new Parameter object and passing
        the args/kwargs to its initialization.
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "parameters"

    def __setattr__(self, name, value) -> None:
        changed = False
        if hasattr(self, "items"):
            if name in self.items:
                if isinstance(value, Parameter):
                    # New parameter object
                    changed = True
                    self.items[name] = value
                elif isinstance(value, str):
                    # A new expression
                    if self.items[name]["value"] != value:
                        changed = True
                        self.items[name]["value"] = value
                        self.items[name].write_expr = True
                else:
                    try:
                        # try a new value, we need to make sure
                        # to stop printing out the expression
                        value = float(value)
                        if self.items[name]["value"] != value:
                            changed = True
                            self.items[name]["value"] = value
                            self.items[name].write_expr = False
                    except:
                        print(
                            "can't set parameter {} to {}".format(
                                self.items[name]["name"], value
                            )
                        )
                if changed:
                    self._changes[name] = value
                    self.__dict__[name] = value
            else:
                self.__dict__[name] = value
        else:
            self.__dict__[name] = value

    def add_parameter(self, *args, **kwargs) -> None:
        p = Parameter(*args, **kwargs)
        self.add_item((p.name, p))


class CompartmentBlock(ModelBlock):
    """
    Compartment block object, subclass of ModelBlock.

    Methods
    -------
    add_compartment(name, dim, size, outside=None)
        adds a compartment by making a new Compartment object and passing
        the args/kwargs to its initialization.
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "compartments"

    def __setattr__(self, name, value) -> None:
        changed = False
        if hasattr(self, "items"):
            if name in self.items:
                if isinstance(value, Compartment):
                    changed = True
                    self.items[name] = value
                elif isinstance(value, str):
                    if self.items[name]["name"] != value:
                        changed = True
                        self.items[name]["name"] = value
                else:
                    try:
                        value = float(value)
                        if self.items[name]["size"] != value:
                            changed = True
                            self.items[name]["size"] = value
                    except:
                        print(
                            "can't set compartment {} to {}".format(
                                self.items[name]["name"], value
                            )
                        )
                if changed:
                    self._changes[name] = value
                    self.__dict__[name] = value
            else:
                self.__dict__[name] = value
        else:
            self.__dict__[name] = value

    def add_compartment(self, *args, **kwargs) -> None:
        c = Compartment(*args, **kwargs)
        self.add_item((c.name, c))


class ObservableBlock(ModelBlock):
    """
    Observable block object, subclass of ModelBlock.

    Methods
    -------
    add_observable(name, otype, patterns=[])
        adds an observable by making a new Observable object and passing
        the args/kwargs to its initialization.
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "observables"

    def __setattr__(self, name, value) -> None:
        changed = False
        if hasattr(self, "items"):
            if name in self.items:
                if isinstance(value, Observable):
                    changed = True
                    self.items[name] = value
                elif isinstance(value, str):
                    if self.items[name]["name"] != value:
                        changed = True
                        self.items[name]["name"] = value
                else:
                    print(
                        "can't set observable {} to {}".format(
                            self.items[name]["name"], value
                        )
                    )
                if changed:
                    self._changes[name] = value
                    self.__dict__[name] = value
            else:
                self.__dict__[name] = value
        else:
            self.__dict__[name] = value

    def add_observable(self, *args, **kwargs) -> None:
        o = Observable(*args, **kwargs)
        self.add_item((o.name, o))


class SpeciesBlock(ModelBlock):
    """
    Species block object, subclass of ModelBlock.

    Methods
    -------
    add_species(name, pattern=Pattern(), count=0)
        adds a species by making a new Species object and passing
        the args/kwargs to its initialization.
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "species"

    def __setattr__(self, name, value) -> None:
        changed = False
        if hasattr(self, "items"):
            if name in self.items:
                if isinstance(value, Species):
                    changed = True
                    self.items[name] = value
                elif isinstance(value, str):
                    if self.items[name]["name"] != value:
                        changed = True
                        self.items[name]["name"] = value
                else:
                    print(
                        "can't set species {} to {}".format(
                            self.items[name]["name"], value
                        )
                    )
                if changed:
                    self._changes[name] = value
                    self.__dict__[name] = value
            else:
                self.__dict__[name] = value
        else:
            self.__dict__[name] = value

    def __getitem__(self, key):
        return self.items[key]

    def __setitem__(self, key, value) -> None:
        self.items[key] = value

    def add_species(self, *args, **kwargs) -> None:
        s = Species(*args, **kwargs)
        ctr = len(self.items)
        self.add_item((ctr, s))


class MoleculeTypeBlock(ModelBlock):
    """
    Molecule type block, subclass of ModelBlock.

    Methods
    -------
    add_molecule_type(name, name, components)
        adds a molecule type by making a new MoleculeType object and passing
        the args/kwargs to its initialization.
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "molecule types"

    def __setattr__(self, name, value) -> None:
        changed = False
        if hasattr(self, "items"):
            if name in self.items:
                if isinstance(value, MoleculeType):
                    changed = True
                    self.items[name] = value
                elif isinstance(value, str):
                    if self.items[name]["name"] != value:
                        changed = True
                        self.items[name]["name"] = value
                else:
                    print(
                        "can't set molecule type {} to {}".format(
                            self.items[name]["name"], value
                        )
                    )
                if changed:
                    self._changes[name] = value
                    self.__dict__[name] = value
            else:
                self.__dict__[name] = value
        else:
            self.__dict__[name] = value

    def add_molecule_type(self, name, components) -> None:
        mt = MoleculeType(name=name, components=components)
        self.add_item((name, mt))


class FunctionBlock(ModelBlock):
    """
    Function block object, subclass of ModelBlock.

    Methods
    -------
    add_function(name, name, expr, args=None)
        adds a function by making a new Function object and passing
        the args/kwargs to its initialization.
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "functions"

    def __setattr__(self, name, value) -> None:
        changed = False
        if hasattr(self, "items"):
            if name in self.items:
                if isinstance(value, Function):
                    changed = True
                    self.items[name] = value
                elif isinstance(value, str):
                    if self.items[name]["expr"] != value:
                        changed = True
                        self.items[name]["expr"] = value
                else:
                    print(
                        "can't set function {} to {}".format(
                            self.items[name]["name"], value
                        )
                    )
                if changed:
                    self._changes[name] = value
                    self.__dict__[name] = value
            else:
                self.__dict__[name] = value
        else:
            self.__dict__[name] = value

    def add_function(self, *args, **kwargs) -> None:
        f = Function(*args, **kwargs)
        self.add_item((f.name, f))


class RuleBlock(ModelBlock):
    """
    Rule block object, subclass of ModelBlock.

    Methods
    -------
    add_rule(name, name, reactants=[], products=[], rate_constants=())
        adds a rule by making a new Rule object and passing
        the args/kwargs to its initialization.
    consolidate_rules : None
        XML loading makes it so that reversible rules are split
        into two unidirectional rules. This find them and combines
        them into a single rule to correctly represent the original
        model rule.
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "reaction rules"

    def __setattr__(self, name, value) -> None:
        changed = False
        if hasattr(self, "items"):
            if name in self.items:
                if isinstance(value, Rule):
                    changed = True
                    self.items[name] = value
                elif isinstance(value, str):
                    if self.items[name]["name"] != value:
                        changed = True
                        self.items[name]["name"] = value
                else:
                    print(
                        "can't set rule {} to {}".format(
                            self.items[name]["name"], value
                        )
                    )
                if changed:
                    self._changes[name] = value
                    self.__dict__[name] = value
            else:
                self.__dict__[name] = value
        else:
            self.__dict__[name] = value

    def add_rule(self, *args, **kwargs) -> None:
        r = Rule(*args, **kwargs)
        self.add_item((r.name, r))

    def consolidate_rules(self) -> None:
        """
        Generated XML only has unidirectional rules
        and uses "_reverse_" tag to make bidirectional
        rules for NFSim. Take all the "_reverse_" tagged
        rules and convert them to bidirectional rules
        """
        delete_list = []
        for item_key in self.items:
            rxn_obj = self.items[item_key]
            if item_key.startswith("_reverse_"):
                # this is the reverse of another reaction
                reverse_of = item_key.replace("_reverse_", "")
                # ensure we have the original
                if reverse_of in self.items:
                    # make bidirectional and add rate law
                    r1 = self.items[reverse_of].rate_constants[0]
                    r2 = rxn_obj.rate_constants[0]
                    self.items[reverse_of].set_rate_constants((r1, r2))
                    # mark reverse for deletion
                    delete_list.append(item_key)
        # delete items marked for deletion
        for del_item in delete_list:
            self.items.pop(del_item)


class ActionBlock(ModelBlock):
    """
    Action block object, subclass of ModelBlock.This is the one object
    that doesn't need a begin/end block tag to be in the model.

    Attributes
    ----------
    _action_list : list[str]
        list of supported action names

    Methods
    -------
    add_action(name, action_type=None, action_args={})
        adds an action by making a new Action object and passing
        the args/kwargs to its initialization.
    clear_actions()
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "actions"
        self.AList = ActionList()
        self._action_list = self.AList.possible_types
        self.items = []
        self.before_model = []

    def __setattr__(self, name, value) -> None:
        self.__dict__[name] = value

    def add_item(self, item_tpl) -> None:
        name, value = item_tpl
        # check to see if it's a before model action
        if self.AList.is_before_model(name):
            self.before_model.append(value)
        else:
            # set the line
            self.items.append(value)

    def __repr__(self) -> str:
        # overwrites what the class representation
        # shows the items in the model block in
        # say ipython
        repr_str = "{} block with {} item(s): {}".format(
            self.name, len(self.items), self.items
        )
        return repr_str

    def __getitem__(self, key):
        return self.items[key]

    def __setitem__(self, key, value) -> None:
        self.items[key] = value

    def __delitem__(self, key) -> None:
        try:
            return self.items.pop(key)
        # TODO: more specific except statements
        except:
            print("Item {} not found".format(key))

    def __iter__(self):
        return self.items.__iter__()

    def __contains__(self, key) -> bool:
        return (key in self.items) or (key in [x.name for x in self.items])

    def add_action(self, action_type, action_args) -> None:
        """
        adds action, needs type as string and args as list of tuples
        (which preserve order) of (argument, value) pairs
        """
        if action_type in self._action_list:
            a = Action(action_type=action_type, action_args=action_args)
            self.add_item((action_type, a))
        else:
            print("Action type {} not valid".format(action_type))

    def clear_actions(self) -> None:
        self.items.clear()

    def gen_string(self) -> str:
        block_lines = []
        # we just loop over lines for actions
        for item in self.items:
            block_lines.append(item.print_line())
        # join everything with new lines
        return "\n".join(block_lines)



import networkx as nx
import stateTransitionDiagram as std
import argparse
import math
import StringIO


def ato_write_gml(graph, fileName, labelGraphics):
    def writeDict(gml, key, label, contents, space, labelGraphics=None):
        gml.write("{1}{0} [\n".format(key, space))
        for subKey in contents:
            if type(contents[subKey]) in [str]:
                gml.write('{2}\t{0} "{1}"\n'.format(subKey, contents[subKey], space))
            elif type(contents[subKey]) in [int]:
                gml.write("{2}\t{0} {1}\n".format(subKey, contents[subKey], space))
            elif type(contents[subKey]) in [dict]:
                writeDict(gml, subKey, subKey, contents[subKey], space + "\t")
            if labelGraphics and label in labelGraphics:
                for labelInstance in labelGraphics[label]:
                    writeDict(
                        gml,
                        "LabelGraphics",
                        "LabelGraphics",
                        labelInstance,
                        space + "\t",
                    )
        gml.write("{0}]\n".format(space))

    gml = StringIO.StringIO()
    gml.write("graph [\n")
    gml.write("\tdirected 1\n")
    for node in graph.node:

        writeDict(gml, "node", node, graph.node[node], "\t", labelGraphics)

    flag = False
    for x in nx.generate_gml(graph):
        if "edge" in x and not flag:
            flag = True
        if flag:
            gml.write(x + "\n")

    # gml.write(']\n')
    with open(fileName, "w") as f:
        f.write(gml.getvalue())
    nx.read_gml(fileName)


def defineConsole():
    """
    defines the program console line commands
    """
    parser = argparse.ArgumentParser(description="SBML to BNGL translator")
    parser.add_argument("-i", "--input", type=str, help="settings file", required=True)
    return parser


def getCounter():
    if not hasattr(getCounter, "counter"):
        getCounter.counter = 0

    getCounter.counter += 1
    return getCounter.counter


def createNode(graph, name, graphicsDict, labelGraphicsDict, isGroup, gid):
    idNumber = getCounter()
    if isGroup:
        if gid:
            graph.add_node(
                name,
                graphics=graphicsDict,
                LabelGraphics=labelGraphicsDict,
                isGroup=isGroup,
                gid=gid,
                id=idNumber,
            )
        else:
            graph.add_node(
                name,
                graphics=graphicsDict,
                LabelGraphics=labelGraphicsDict,
                isGroup=isGroup,
                id=idNumber,
            )
    else:
        graph.add_node(
            name,
            graphics=graphicsDict,
            LabelGraphics=labelGraphicsDict,
            gid=gid,
            id=idNumber,
        )
    return idNumber


def createBitNode(graph, molecule, nodeList, simplifiedText):
    """
    creates the bit nodes
    """
    gridDict = {
        1: 1,
        2: 2,
        3: 3,
        4: 2,
        5: 3,
        6: 3,
        7: 4,
        8: 4,
        9: 3,
        10: 5,
        11: 4,
        12: 4,
        13: 5,
        14: 5,
        15: 5,
        16: 4,
        17: 5,
        18: 4,
        19: 5,
        20: 5,
    }

    for node in nodeList[molecule]:
        componentLegend = ""
        nodeName = u""
        nodeId = []
        for idx, bit in enumerate(node):
            componentLegend += bit[0]
            if bit[1]:
                if simplifiedText:
                    nodeName += "o"
                else:
                    nodeName += u"\u25CF "
                nodeId.append(bit[0])
            else:
                if simplifiedText:
                    nodeName += "x"
                else:
                    nodeName += u"\u25CB "
                # nodeName += u"\u00B7 "
            if (idx + 1) % gridDict[len(node)] == 0 and idx + 1 != len(node):
                nodeName.strip(" ")
                nodeName += "\n"
                componentLegend += "\n"
            else:
                componentLegend += "/"
        createNode(
            graph,
            "{0}_{1}".format(molecule, "/".join(nodeId)),
            {"type": "roundrectangle"},
            {"text": nodeName},
            0,
            graph.node[molecule]["id"],
        )
    createNode(
        graph,
        "{0}_legend".format(molecule),
        {},
        {"text": componentLegend, "fontSize": 20},
        0,
        graph.node[molecule]["id"],
    )


def createBitEdge(graph, molecule, edge):
    nodeId0 = [x[0] for x in edge[0] if x[1]]
    nodeId1 = [x[0] for x in edge[1] if x[1]]
    bidirectional = (edge[1], edge[0]) in edgeList[molecule]
    if (
        "{0}_{1}".format(molecule, "/".join(nodeId1)),
        "{0}_{1}".format(molecule, "/".join(nodeId0)),
    ) not in graph.edges():
        if bidirectional:
            graph.add_edge(
                "{0}_{1}".format(molecule, "/".join(nodeId0)),
                "{0}_{1}".format(molecule, "/".join(nodeId1)),
                graphics={
                    "fill": "#000000",
                    "sourceArrow": "standard",
                    "targetArrow": "standard",
                },
            )

        else:
            graph.add_edge(
                "{0}_{1}".format(molecule, "/".join(nodeId0)),
                "{0}_{1}".format(molecule, "/".join(nodeId1)),
                graphics={"fill": "#000000", "targetArrow": "standard"},
            )


def createPDLabelNode(graph, molecule, nodeList):
    globalLabelDict = {}
    for node in nodeList[molecule]:
        localNodeName = "{0}_{1}".format(
            molecule, "/".join([x[0] for x in node if x[1]])
        )
        labelDict = []
        anchorArray = ["c", "t", "tl", "tr", "l", "r", "bl", "b", "br"]
        for anchor, bit in zip(anchorArray, node):
            localLabelDict = {}

            # localLabelDict['outline'] = '#FFFFFF'
            localLabelDict["text"] = str(bit[0])
            localLabelDict["fill"] = "#FF88FF" if bit[1] else "#FFFFFF"
            localLabelDict["fontSize"] = 24
            localLabelDict["anchor"] = anchor

            labelDict.append(localLabelDict)

        createNode(
            graph,
            localNodeName,
            {"type": "roundrectangle", "fill": "#FFFFFF", "w": 270, "h": 140},
            {},
            0,
            graph.node[molecule]["id"],
        )
        globalLabelDict[localNodeName] = labelDict
    return globalLabelDict


def createPDGraphNode(graph, molecule, nodeList):
    for node in nodeList[molecule]:
        localNodeName = "{0}_{1}".format(
            molecule, "/".join([x[0] for x in node if x[1]])
        )
        createNode(
            graph,
            localNodeName,
            {"type": "roundrectangle", "fill": "#FFFFFF"},
            {"text": " "},
            1,
            graph.node[molecule]["id"],
        )

        localNodeid = graph.node[localNodeName]["id"]
        subNodeId = []
        for bit in node:
            if bit[1]:
                nodecolor = "#FF88FF"
                createNode(
                    graph,
                    "{0}_{1}".format(localNodeid, bit[0]),
                    {"type": "roundrectangle", "fill": nodecolor},
                    {"text": bit[0]},
                    0,
                    localNodeid,
                )

            else:
                nodecolor = "#FFFFFF"
                createNode(
                    graph,
                    "{0}_{1}".format(localNodeid, bit[0]),
                    {"type": "roundrectangle", "fill": nodecolor},
                    {"text": bit[0]},
                    0,
                    localNodeid,
                )
            subNodeId.append("{0}_{1}".format(localNodeid, bit[0]))

        for idx in range(0, len(subNodeId) - 1):
            graph.add_edge(
                subNodeId[idx],
                subNodeId[idx + 1],
                graphics={"fill": "#CCCCFF", "style": "dashed"},
            )

        # for idx in range(0, len(subNodeId)/2):
        #    graph.add_edge(subNodeId[idx], subNodeId[idx+len(subNodeId)/2], graphics={'fill': '#CCCCFF', 'style': 'dashed'})


def createPDEdge(graph, molecule, edge, edgeList):
    nodeId0 = [x[0] for x in edge[0] if x[1]]
    nodeId1 = [x[0] for x in edge[1] if x[1]]

    bidirectionalArray = [
        x for x in edgeList[molecule] if (edge[1], edge[0]) == (x[0], x[1])
    ]

    bidirectional = len(bidirectionalArray) > 0

    if (
        "{0}_{1}".format(molecule, "/".join(nodeId1)),
        "{0}_{1}".format(molecule, "/".join(nodeId0)),
    ) not in graph.edges():
        transDict = {False: "^", True: ""}
        processNodeName = "{0}_{1}_{2}_edge".format(
            molecule, "/".join(nodeId0), "/".join(nodeId1)
        )
        if "reverse" in edge[2]:
            difference = set(edge[0]) - set(edge[1])
        else:
            difference = set(edge[1]) - set(edge[0])

        # if not bidirectional:
        #    differenceText = ['{1}{0}'.format(x[0],transDict[x[1]]) for x in difference]
        # else:
        #    differenceText = ['{0}'.format(x[0]) for x in difference]

        differencePositive = [x[0] for x in difference if x[1]]
        differenceNegative = [x[0] for x in difference if not x[1]]

        if bidirectional:
            if "reverse" in edge[2]:
                differenceText = bidirectionalArray[0][3] + ", " + edge[3]
            else:
                differenceText = edge[3] + ", " + bidirectionalArray[0][3]
        else:
            differenceText = edge[3]

        # differenceText = edge[2].strip('_reverse_')
        differenceText += "\n===\n"
        differenceText = ""

        # if bidirectional and len(differencePositive) < len(differenceNegative):
        #    differencePositive, differenceNegative = differenceNegative, differencePositive
        if len(differencePositive) > 0 and len(differenceNegative) > 0:
            differenceText += (
                ", ".join(differencePositive)
                + "\n___\n"
                + ", ".join(["^{0}".format(x) for x in differenceNegative])
            )
        else:
            differenceText += ", ".join(differencePositive) + ", ".join(
                ["^{0}".format(x) for x in differenceNegative]
            )

        # creates the process nodes with text 'differenceText'
        createNode(
            graph,
            processNodeName,
            {"type": "rectangle", "fill": "#FFFFFF"},
            {"text": differenceText, "fontStyle": "bold", "fontSize": 20},
            0,
            graph.node[molecule]["id"],
        )
        if bidirectional:
            graph.add_edge(
                "{0}_{1}".format(molecule, "/".join(nodeId0)),
                processNodeName,
                graphics={"fill": "#000000", "sourceArrow": "standard", "width": 3},
            )
            graph.add_edge(
                processNodeName,
                "{0}_{1}".format(molecule, "/".join(nodeId1)),
                graphics={"fill": "#000000", "targetArrow": "standard", "width": 3},
            )
        else:
            graph.add_edge(
                "{0}_{1}".format(molecule, "/".join(nodeId0)),
                processNodeName,
                graphics={"fill": "#000000", "width": 3},
            )
            graph.add_edge(
                processNodeName,
                "{0}_{1}".format(molecule, "/".join(nodeId1)),
                graphics={"fill": "#000000", "targetArrow": "standard", "width": 3},
            )
    return bidirectional


def generateSTD(nodeList, edgeList, simplifiedText=False):
    graph = nx.DiGraph()
    globalLabelDict = {}
    for molecule in nodeList:
        createNode(
            graph,
            molecule,
            {"type": "roundrectangle", "fill": "#FFFFFF"},
            {
                "fontSize": 24,
                "fontStyle": "bold",
                "alignment": "right",
                "autoSizePolicy": "node_size",
            },
            1,
            0,
        )

        # globalLabelDict.update(createPDLabelNode(graph, molecule, nodeList))
        createBitNode(graph, molecule, nodeList, simplifiedText)

    for molecule in nodeList:

        bidirectionalList = []
        cedgeList = list(edgeList[molecule])
        for edge in cedgeList:
            if "_reverse" in edge[2]:
                continue
            if list([edge[0], edge[1]]) not in bidirectionalList:

                bidirectional = createPDEdge(graph, molecule, edge, edgeList)
            if bidirectional:
                bidirectionalList.append([edge[1], edge[0]])

    return graph


def outputGraph(graph, fileName, labelDict):
    nx.write_gml(graph, fileName)
    # ato_write_gml(graph, fileName, labelDict)


import codecs


def generateSTDGML(inputFile, simplifiedText=False):
    nodeList, edgeList = std.getContextRequirements(inputFile, excludeReverse=False)
    graph = generateSTD(nodeList, edgeList, simplifiedText)
    # graph = nx.convert_node_labels_to_integers(graph)
    # gml = nx.generate_gml(graph)
    # print graph
    # nx.write_gml(graph,inputFile+'_std.gml')
    # with open(inputFile+'_std.gml','r') as f:
    #    gml = f.read()
    # gml = codecs.open(inputFile + '.gml', 'r','utf-8')
    # gml = ''.join(gml)
    # return gml.read()
    # return graph
    # return ''.join(gml)
    # gmlstr =  ''.join(gml)
    # gmlstr = gmlstr.encode('utf-8')
    return graph


if __name__ == "__main__":

    parser = defineConsole()
    namespace = parser.parse_args()
    inputFile = namespace.input

    graph = generateSTDGML(inputFile)
    nx.write_gml(graph, inputFile + "_std.gml")
    # print gmlgraph[0:400]
    # nxgml = nx.parse_gml(gmlgraph)
    # import gml2cyjson
    # gmlgraph = nx.parse_gml(gmlgraph)
    # gml2cyjson.gml2cyjson(gmlgraph,'std')
    # nodeList, edgeList = std.getContextRequirements(inputFile, excludeReverse=True)
    # graph = generateSTD(nodeList, edgeList)
    # outputGraph(graph, '{0}_std.gml'.format(namespace.input), {})



import xmltodict, re

from bionetgen.main import BioNetGen
from tempfile import TemporaryFile

from .bngfile import BNGFile
from .xmlparsers import ParameterBlockXML, CompartmentBlockXML, ObservableBlockXML
from .xmlparsers import SpeciesBlockXML, MoleculeTypeBlockXML, FunctionBlockXML
from .xmlparsers import RuleBlockXML
from .blocks import ActionBlock

# This allows access to the CLIs config setup
app = BioNetGen()
app.setup()
conf = app.config["bionetgen"]
def_bng_path = conf["bngpath"]


class BNGParser:
    """
    Parser object that deals with reading in the BNGL file and
    setting up the model object

    Usage: BNGParser(bngl_path)
           BNGParser(bngl_path, BNGPATH)

    Attributes
    ----------
    bngfile : BNGFile
        BNGFile object that's responsible for .bngl file manipulations

    Methods
    -------
    parse_model(model_file)
        parses the BNGL model at the given path and adds everything to a given model object
    parse_xml(xml_str)
        parses given xml string and adds everything to a given model object
    """

    def __init__(self, path, BNGPATH=def_bng_path, parse_actions=True) -> None:
        self.to_parse_actions = parse_actions
        self.bngfile = BNGFile(path)

    def parse_model(self, model_obj) -> None:
        """
        Will determine the parser route eventually and call the right
        parser
        """
        self._parse_model_bngpl(model_obj)
        if self.to_parse_actions:
            self.parse_actions(model_obj)

    def _parse_model_bngpl(self, model_obj) -> None:
        # get file path
        model_file = self.bngfile.path

        # this route runs BNG2.pl on the bngl and parses
        # the XML instead
        if model_file.endswith(".bngl"):
            # TODO: Add verbosity option to the library
            # print("Attempting to generate XML")
            with TemporaryFile("w+") as xml_file:
                if self.bngfile.generate_xml(xml_file):
                    # TODO: Add verbosity option to the library
                    xmlstr = xml_file.read()
                    # < is not a valid XML character, we need to replace it
                    xmlstr = xmlstr.replace('relation="<', 'relation="&lt;')
                    self.parse_xml(xmlstr, model_obj)
                    model_obj.reset_compilation_tags()
                else:
                    raise ValueError("XML file couldn't be generated")
        elif model_file.endswith(".xml"):
            with open(model_file, "r") as f:
                xml_str = f.read()
                # < is not a valid XML character, we need to replace it
                xmlstr = xml_str.replace('relation="<', 'relation="&lt;')
                self.parse_xml(xml_str, model_obj)
            model_obj.reset_compilation_tags()
        else:
            raise NotImplementedError(
                "The extension of {} is not supported".format(model_file)
            )

    # TODO: This requires a massive clean up. Might need
    # to make a grammar for it.
    def parse_actions(self, model_obj):
        if len(self.bngfile.parsed_actions) > 0:
            ablock = ActionBlock()
            # we have actions in file, let's get them
            for action in self.bngfile.parsed_actions:
                action = re.sub(
                    r"\#.*", "", action
                )  # should this be (r"\#.*) or just ("\#.*")
                action = re.sub(
                    r"\s", "", action
                )  # should this be (r"\s) or just ("\s")
                if len(action) == 0:
                    continue
                # gotta find if actions argument is given
                actions_arg = None
                amatch = re.match(r".*(actions=>\[(.*)\]).*", action)
                if amatch is not None:
                    # we have an actions argument, let's put that aside for now
                    actions_arg_match = re.match(r".*\[(.*)\].*", amatch.group(1))
                    if actions_arg_match is not None:
                        actions_arg = actions_arg_match.group(1)
                    # let's remove the action argument to allow regular parsing
                    action = action.replace(amatch.group(1), "")
                    # let's make sure we don't have two commas back to back now
                    commatch = re.match(r".*(\,\s*\,).*", action)
                    if commatch is not None:
                        action = action.replace(commatch.group(1), ",")
                m = re.match(r"^\s*(\S+)\(\s*(\S*)\s*\)(\;)?\s*(\#\s*\S*)?\s*", action)
                if m is not None:
                    # here we have an action
                    atype = m.group(1)
                    in_parens = m.group(2)
                    if len(in_parens) > 0:
                        # in parenthesis group can have curly or square braces
                        m = re.match(r"\{(\S*)\}$", in_parens)
                        arg_dict = {}
                        if actions_arg is not None:
                            arg_dict["actions"] = f"[{actions_arg}]"
                        if m is not None:
                            arg_list_str = m.group(1)
                            # First let's check for lists
                            # Please note that this will only replace a single list that doesn't reoccur
                            L = re.match(r"\S+\[(\S*)\]\S*", m.group(1))
                            if L is not None:
                                arg_list_str = arg_list_str.replace(
                                    f"[{L.group(1)}]",
                                    f"[{L.group(1).replace(',','_')}]",
                                )
                            # Now check for curly braces
                            # Please note that this will only replace a single dictionary that doesn't reoccur
                            L = re.match(r"\S+\{(\S*)\}\S*", m.group(1))
                            if L is not None:
                                arg_list_str = arg_list_str.replace(
                                    "{%s}" % L.group(1),
                                    "{%s}" % L.group(1).replace(",", "_"),
                                )
                            arg_list = arg_list_str.split(",")
                            for arg_str in arg_list:
                                splt = arg_str.split("=>")
                                if len(splt) > 1:
                                    arg = splt[0]
                                    val = "=>".join(splt[1:])
                                    # now we need to check if we have a list/dictionary
                                    if "_" in val:
                                        val = val.replace("_", ",")
                                    # add the tuple to the list
                                    if arg in arg_dict:
                                        # TODO: make this a warning
                                        print(
                                            f"WARNING: argument {arg} for action {atype} is given twice, using the last value {val}"
                                        )
                                    arg_dict[arg] = val
                        else:
                            m = re.match(r"\[(\S*)\]", in_parens)
                            if m is not None:
                                # this is a list of items
                                arg_list_str = m.group(1)
                            else:
                                # what we have in parentheses has to
                                # be a simple list of arguments
                                arg_list_str = in_parens
                            arg_list = arg_list_str.split(",")
                            for arg_str in arg_list:
                                # add to arg_tuples
                                if arg_str in arg_dict:
                                    # TODO: make this a warning
                                    print(
                                        f"WARNING: argument {arg_str} for action {atype} is given twice"
                                    )
                                arg_dict[arg_str] = None
                        ablock.add_action(atype, arg_dict)
                    else:
                        ablock.add_action(atype, {})
            model_obj.add_block(ablock)

    def parse_xml(self, xml_str, model_obj) -> None:
        xml_dict = xmltodict.parse(xml_str)
        model_obj.xml_dict = xml_dict
        xml_model = xml_dict["sbml"]["model"]
        model_obj.model_name = xml_model["@id"]
        for listkey in xml_model.keys():
            if listkey == "ListOfParameters":
                param_list = xml_model[listkey]
                if param_list is not None:
                    params = param_list["Parameter"]
                    xml_parser = ParameterBlockXML(params)
                    model_obj.add_block(xml_parser.parsed_obj)
            elif listkey == "ListOfObservables":
                obs_list = xml_model[listkey]
                if obs_list is not None:
                    obs = obs_list["Observable"]
                    xml_parser = ObservableBlockXML(obs)
                    model_obj.add_block(xml_parser.parsed_obj)
            elif listkey == "ListOfCompartments":
                comp_list = xml_model[listkey]
                if comp_list is not None:
                    comps = comp_list["compartment"]
                    xml_parser = CompartmentBlockXML(comps)
                    model_obj.add_block(xml_parser.parsed_obj)
            elif listkey == "ListOfMoleculeTypes":
                mtypes_list = xml_model[listkey]
                if mtypes_list is not None:
                    mtypes = mtypes_list["MoleculeType"]
                    xml_parser = MoleculeTypeBlockXML(mtypes)
                    model_obj.add_block(xml_parser.parsed_obj)
            elif listkey == "ListOfSpecies":
                species_list = xml_model[listkey]
                if species_list is not None:
                    species = species_list["Species"]
                    xml_parser = SpeciesBlockXML(species)
                    model_obj.add_block(xml_parser.parsed_obj)
            elif listkey == "ListOfReactionRules":
                rrules_list = xml_model[listkey]
                if rrules_list is not None:
                    rrules = rrules_list["ReactionRule"]
                    xml_parser = RuleBlockXML(rrules)
                    model_obj.add_block(xml_parser.parsed_obj)
            elif listkey == "ListOfFunctions":
                # TODO: Optional expression parsing?
                # TODO: Add arguments correctly
                func_list = xml_model[listkey]
                if func_list is not None:
                    funcs = func_list["Function"]
                    xml_parser = FunctionBlockXML(funcs)
                    model_obj.add_block(xml_parser.parsed_obj)
        # And that's the end of parsing
        # TODO: Add verbosity option to the library
        # print("Parsing complete")



# -*- coding: utf-8 -*-
"""
Created on Sat Oct 19 15:19:35 2013

@author: proto
"""
import pprint
import difflib
from collections import Counter
import json
import ast
import pickle
import os
from os import listdir
from os.path import isfile, join
import numpy as np

try:
    from utils.util import pmemoize as memoize
    import libsbml
except ModuleNotFoundError:
    import sys

    sys.path.append("..")
    from bionetgen.atomizer.utils.util import pmemoize as memoize
    import libsbml


@memoize
def levenshtein(s1, s2):
    l1 = len(s1)
    l2 = len(s2)

    # matrix = [range(l1 + 1)] * (l2 + 1)
    # for zz in range(l2 + 1):
    #  matrix[zz] = range(zz,zz + l1 + 1)
    matrix = [list(range(x, x + l1 + 1)) for x in range(0, l2 + 1)]
    for zz in range(0, l2):
        for sz in range(0, l1):
            z = matrix[zz][sz] if s1[sz] == s2[zz] else matrix[zz][sz] + 1
            matrix[zz + 1][sz + 1] = min(
                matrix[zz + 1][sz] + 1, matrix[zz][sz + 1] + 1, z
            )
    return matrix[l2][l1]


def getDifferences(scoreMatrix, speciesName, threshold):
    """
    given a list of strings and a scoreMatrix, return the list of difference between
    those strings with a levenshtein difference of less than threshold
    returns:
        namePairs: list of tuples containing strings with distance <2
        differenceList: list of differences between the tuples in namePairs
    """
    differenceList = []
    namePairs = []

    for idx in range(0, len(scoreMatrix)):
        for idx2 in range(0, len(scoreMatrix[idx])):
            if scoreMatrix[idx][idx2] <= threshold and idx < idx2:
                if len(speciesName[idx]) < len(speciesName[idx2]):
                    namePairs.append([speciesName[idx], speciesName[idx2]])
                else:
                    namePairs.append([speciesName[idx2], speciesName[idx]])
    for pair in namePairs:
        if len(pair[1]) < len(pair[0]):
            difference = difflib.ndiff(pair[1].lower(), pair[0].lower())
        else:
            difference = difflib.ndiff(pair[0].lower(), pair[1].lower())
        tmp = []
        for diff in difference:
            if diff[0] == " ":
                continue
            tmp.append(diff)
        differenceList.append(tuple(tmp))

    return namePairs, differenceList


def loadOntology(ontologyFile):
    if os.path.isfile(ontologyFile):
        tmp = {}
        with open(ontologyFile, "r") as fp:
            ontology = json.load(fp)
        for element in ontology["patterns"]:
            tmp[ast.literal_eval(element)] = ontology["patterns"][element]
        ontology["patterns"] = tmp
        return ontology
    else:
        tmp = {}
        ontology = {
            "modificationList": ["Phosporylation", "Double-Phosporylation"],
            "reactionSite": ["phospho"],
            "reactionState": ["P", "PP"],
            "definitions": [{"rsi": 0, "rst": 0}, {"rsi": 0, "rst": 1}],
            "patterns": {
                "('+ _', '+ P', '+ P')": "Double-Phosporylation",
                "('+ P', '+ P', '+ _')": "Double-Phosporylation",
                "('+ p', '+ p')": "Double-Phosporylation",
            },
        }
        for element in ontology["patterns"]:
            tmp[ast.literal_eval(element)] = ontology["patterns"][element]
        ontology["patterns"] = tmp
        return ontology


def findLongestSubstring(speciesA, speciesB):
    sm = difflib.SequenceMatcher(a=speciesA, b=speciesB)
    longestMatch = sm.find_longest_match(0, len(speciesA), 0, len(speciesB))

    longestMatchStr = speciesA[longestMatch[0] : longestMatch[0] + longestMatch[2]]
    # trim around '_' separators for more natural name cuts
    trimIndex = [0, 0]
    _index = [i for i, x in enumerate(speciesA) if x == "_"]
    if "_" in speciesA:
        if not any([longestMatch[0] + longestMatch[2] == x for x in _index]):
            for idx in range(0, len(_index)):
                if idx + 1 == len(_index):
                    trimIndex[1] = _index[idx]
                elif idx == 0:
                    continue
                elif _index[idx] - longestMatch[0] > 0:
                    trimIndex[1] = _index[idx - 1]
            longestMatchStr = speciesA[trimIndex[0] : trimIndex[1]]

    return longestMatchStr


# import concurrent.futures


def stringToSet(species, idx, scoreRow, speciesName):
    for idx2, species2 in enumerate(speciesName):
        if species == species2 or scoreRow[idx2] != 0:
            continue
        scoreRow[idx2] = levenshtein(species, speciesName[idx2])
    return idx, scoreRow


def defineEditDistanceMatrix3(speciesName, similarityThreshold=4, parallel=False):
    namePairs = []
    differenceList = []
    for species in speciesName:
        closeMatches = difflib.get_close_matches(species, speciesName)
        closeMatches = [x for x in closeMatches if len(x) < len(species)]
        for match in closeMatches:
            difference = [x for x in difflib.ndiff(match, species)]
            if len([x for x in difference if "-" in x]) == 0:
                namePairs.append([match, species])
                differenceList.append(tuple([x for x in difference if "+" in x]))
    return namePairs, differenceList, ""


def defineEditDistanceMatrix(speciesName, similarityThreshold=4, parallel=False):
    """
    obtains a distance matrix and a pairs of elements that are close
    in distance, along with the proposed differences
    """
    differenceCounter = Counter()
    futures = []
    scoreMatrix = np.zeros((len(speciesName), len(speciesName)))
    scoreMatrix2 = np.zeros((len(speciesName), len(speciesName)))
    counter = range(0, len(speciesName))
    """
    for idx,species in enumerate(speciesName):
    
        futures = []
        with concurrent.futures.ProcessPoolExecutor(max_workers=4) as executor:
            futures.append(executor.submit(stringToSet, species,idx,scoreMatrix[idx],speciesName))
        for future in concurrent.futures.as_completed(futures):
            idx3,row = future.result()
            counter.remove(idx3)
            print(len(counter))
            scoreMatrix[idx3] = row
    """

    for idx, species in enumerate(speciesName):
        for idx2, species2 in enumerate(speciesName):
            if species == species2 or scoreMatrix2[idx][idx2] != 0:
                continue
            # comparison = difflib.SequenceMatcher(None,speciesName[idx],speciesName[idx2]).quick_ratio()
            comparison = levenshtein(speciesName[idx], speciesName[idx2])
            scoreMatrix2[idx][idx2] = comparison
            scoreMatrix2[idx2][idx] = scoreMatrix2[idx][idx2]

    namePairs, differenceList = getDifferences(
        scoreMatrix2, speciesName, similarityThreshold
    )

    differenceCounter.update(differenceList)
    return namePairs, differenceList, differenceCounter


def analyzeNamingConventions(
    speciesName, ontologyFile, ontologyDictionary={}, similarityThreshold=4
):
    patternClassification = {}
    pairClassification = {}

    # ontology =  loadOntology(ontologyFile)
    ontology = ontologyFile
    finalDifferenceCounter = Counter()
    namePairs, differenceList, differenceCounter = defineEditDistanceMatrix(
        speciesName, similarityThreshold
    )

    for element in differenceCounter:
        if element in ontology["patterns"]:
            finalDifferenceCounter[element] = differenceCounter[element]
            patternClassification[element] = ontology["patterns"][element]
    for pair, difference in zip(namePairs, differenceList):
        if difference in patternClassification:
            if patternClassification[difference] not in pairClassification:
                pairClassification[patternClassification[difference]] = []
            pairClassification[patternClassification[difference]].append(tuple(pair))
    keys = finalDifferenceCounter.keys()
    tmp = {key: "" for key in keys}
    for key in tmp:
        tmp[key] = ontology["patterns"][key]
    keys = ["".join(x).replace("+ ", "") for x in keys]
    # print(ontology)

    return pairClassification, keys, tmp


def main(fileName):
    reader = libsbml.SBMLReader()
    document = reader.readSBMLFromFile(fileName)
    model = document.getModel()
    speciesName = []

    for species in model.getListOfSpecies():
        speciesName.append(species.getName())

    return analyzeNamingConventions(
        speciesName, "reactionDefinitions/namingConventions.json"
    )


def databaseAnalysis(directory, outputFile):
    xmlFiles = [
        f
        for f in listdir("./" + directory)
        if isfile(join("./" + directory, f)) and f.endswith("xml")
    ]
    differenceCounter = Counter()
    fileDict = {}
    for xml in xmlFiles:
        print(xml)
        reader = libsbml.SBMLReader()
        document = reader.readSBMLFromFile(directory + xml)
        model = document.getModel()
        if model == None:
            continue
        speciesName = []

        for species in model.getListOfSpecies():
            speciesName.append(species.getName())
        scoreMatrix = np.zeros((len(speciesName), len(speciesName)))
        for idx, species in enumerate(speciesName):
            for idx2, species2 in enumerate(speciesName):
                if species == species2 or scoreMatrix[idx][idx2] != 0:
                    continue
                scoreMatrix[idx][idx2] = levenshtein(
                    speciesName[idx], speciesName[idx2]
                )
                scoreMatrix[idx2][idx] = scoreMatrix[idx][idx2]

        namePairs, differenceList = getDifferences(scoreMatrix, speciesName, 3)
        differenceCounter.update(differenceList)
        for key, element in zip(differenceList, namePairs):
            if key == ():
                continue
            if key not in fileDict:
                # differenceDict[tuple(key)] = set()
                fileDict[tuple(key)] = set()
            # differenceDict[tuple(key)].add((xml,tuple(element)))
            fileDict[tuple(key)].add(xml)

        fileCounter = Counter()
        for element in fileDict:
            fileCounter[element] = len(fileDict[element])
        with open(outputFile, "wb") as f:
            pickle.dump(differenceCounter, f)
            # pickle.dump(differenceDict,f)
            pickle.dump(fileCounter, f)


"""        
try:
    import pandas as pd
except ImportError:
    pd = None

def analyzeTrends(inputFile):
    with open(inputFile,'rb') as f:
        counter = pickle.load(f)
        #dictionary = pickle.load(f)
        fileCounter = pickle.load(f)
    totalCounter = Counter()
    for element in counter:
        
        totalCounter[element] = counter[element] * fileCounter[element]/469.0
    keys = totalCounter.most_common(200)
    #keys = keys[1:]
    pp = pprint.PrettyPrinter(indent=4)
    pp.pprint(keys)
    data = pd.DataFrame(keys)
    #print(data.to_excel('name.xls'))
    
    #for element in keys:
    #    print('------------------')
    #    print(element)
    #    pp.pprint(dictionary[element[0]])
"""

if __name__ == "__main__":
    bioNumber = 19
    main("XMLExamples/curated/BIOMD%010i.xml" % bioNumber)

    # databaseAnalysis('XMLExamples/non_curated/', 'non_ontologies.dump')
    # analyzeTrends('ontologies.dump')



import os
from cement.utils.version import get_version as cement_get_version

# Find VERSION file
vpath = os.path.dirname(os.path.abspath(__file__))
vpath = os.path.split(vpath)[0]
vpath = os.path.join(*[vpath, "assets", "VERSION"])
with open(vpath, "r") as f:
    v = f.read()
vtuple = [0, 0, 0, 0, 0]
for iv, ver in enumerate(v.split()):
    try:
        vtuple[iv] = int(ver)
    except:
        vtuple[iv] = ver

VERSION = tuple(vtuple)


def get_version(version=VERSION):
    return cement_get_version(version)



# NOTE Anything that needs to go into the library
# side needs to not be in the core section, it
# leads to circular imports
from .result import BNGResult
from .plot import BNGPlotter
from .info import BNGInfo
from .cli import BNGCLI
from .visualize import BNGVisualize



import bionetgen as bng
from bionetgen.modelapi.utils import run_command
import bionetgen.modelapi.model as mdl
import os, subprocess
from tempfile import NamedTemporaryFile
from bionetgen.tools import BNGPlotter
from bionetgen.tools import BNGInfo
from bionetgen.tools import BNGVisualize
from bionetgen.tools import BNGCLI

import os, sys


# TODO Consolidate how config is being accessed. It's
# almost like each function accesses the configs from
# a different path


def runCLI(config, args):
    """
    Convenience function to run BNG2.pl from the CLI app

    Usage: runCLI(config, args)

    Arguments
    ---------
    config : dict
        configuration dictionary from BioNetGen cement app
    args :  argparse.Namespace
        arguments parsed from the command line with argparser.
    """
    # this pulls out the arguments
    sys.tracebacklimit = args.traceback_depth
    inp_file = args.input
    output = args.output
    log_file = args.log_file
    # if you set args.bngpath it should take precedence
    config_bngpath = config.get("bionetgen", "bngpath")
    # and instantiates the CLI object
    cli = BNGCLI(inp_file, output, config_bngpath, log_file=log_file)
    cli.stdout = config.get("bionetgen", "stdout")
    cli.stderr = config.get("bionetgen", "stderr")
    cli.run()


def plotDAT(inp, out=".", kw=dict()):
    """
    Convenience function to plot dat/scan files from the CLI

    Usage: plotDAT(inp, out, kw)

    Arguments
    ---------
    inp : str
        input gdat/cdat/scan file to plot
    out : str
        (optional) output file path, can be used to define the
        output format as well. Default is the current folder,
        filename is the same as the input file and default format
        is PNG.
    kw : dict
        (optional) this is a set of keyword arguments you want to
        pass for certain matplotlib options. Check -h for details
    """
    # if we want to plot directly into the folder
    # we are in we need to get the path correctly
    if out == ".":
        path, fname = os.path.split(inp)
        fnoext, ext = os.path.splitext(fname)
        out = os.path.join(path, "{}.png".format(fnoext))
    # use the plotter object to get the plot
    from bionetgen.core import BNGPlotter

    plotter = BNGPlotter(inp, out, **kw)
    plotter.plot()


def runAtomizeTool(config, args):
    from bionetgen.atomizer import AtomizeTool

    a = AtomizeTool(parser_namespace=args)
    # do config specific stuff here if need be, or remove the config requirement
    a.run()


def printInfo(config, args):
    """
    Uses BNGInfo class to print BioNetGen information using
    arguments and config from Cement framework.
    """
    # this pulls out the arguments
    # inp_file = args.input
    # output = args.output
    # if you set args.bngpath it should take precedence
    # config_bngpath = config.get("bionetgen", "bngpath")
    # and instantiates the CLI object
    info = BNGInfo(config=config)
    info.gatherInfo()
    info.messageGeneration()
    info.run()


def visualizeModel(config, args):
    """
    Uses BNGVisualize class to visualize BNGL models using
    arguments and configuration from Cement framework.
    """
    inp = args.input
    out = args.output
    vtype = args.type
    # if you set args.bngpath it should take precedence
    config_bngpath = config.get("bionetgen", "bngpath")
    viz = BNGVisualize(inp, output=out, vtype=vtype, bngpath=config_bngpath)
    viz.run()



# -*- coding: utf-8 -*-
"""
Created on Wed May 30 11:44:17 2012

@author: proto
"""
from copy import deepcopy
import difflib
import hashlib
import numpy
from collections import Counter, defaultdict


def compareLists(list1, list2):
    return Counter(list1) == Counter(list2)


class Species:
    def __init__(self):
        self.molecules = []
        self.bondNumbers = []

    def __eq__(self, other):
        return str(self) == str(other)

    def getBondNumbers(self):
        bondNumbers = [0]
        for element in self.molecules:
            bondNumbers.extend(element.getBondNumbers())
        return bondNumbers

    def copy(self):
        species = Species()
        for molecule in self.molecules:
            species.molecules.append(molecule.copy())
        return species

    def addMolecule(self, molecule, concatenate=False, iteration=1):
        if not concatenate:
            self.molecules.append(molecule)
        else:
            counter = 1
            for element in self.molecules:
                if element.name == molecule.name:
                    if iteration == counter:
                        element.extend(molecule)
                        return
                    else:
                        counter += 1
            self.molecules.append(molecule)
            # self.molecules.append(molecule)
            # for element in self.molecules:
            #    if element.name == molecule.name:

    def addCompartment(self, tags):
        for molecule in self.molecules:
            molecule.setCompartment(tags)

    def deleteMolecule(self, moleculeName):
        deadMolecule = None
        for element in self.molecules:
            if element.name == moleculeName:
                deadMolecule = element
        if deadMolecule == None:
            return
        bondNumbers = deadMolecule.getBondNumbers()
        self.molecules.remove(deadMolecule)
        for element in self.molecules:
            for component in element.components:
                for number in bondNumbers:
                    if str(number) in component.bonds:
                        component.bonds.remove(str(number))

    def getMolecule(self, moleculeName):
        for molecule in self.molecules:
            if moleculeName == molecule.name:
                return molecule
        return None

    def getSize(self):
        return len(self.molecules)

    def getMoleculeNames(self):
        return [x.name for x in self.molecules]

    def contains(self, moleculeName):
        for molecule in self.molecules:
            if moleculeName == molecule.name:
                return True
        return False

    def addChunk(self, tags, moleculesComponents, precursors):
        """
        temporary transitional method
        """
        for (tag, components) in zip(tags, moleculesComponents):
            if self.contains(tag):
                tmp = self.getMolecule(tag)
            else:
                tmp = Molecule(tag)
                # for element in precursors:
                #    if element.getMolecule(tag) != None:
                #        tmp = element.getMolecule(tag)

            for component in components:
                if tmp.contains(component[0][0]):
                    tmpCompo = tmp.getComponent(component[0][0])
                    # continue
                else:
                    tmpCompo = Component(component[0][0])

                for index in range(1, len(component[0])):
                    tmpCompo.addState(component[0][index])
                if len(component) > 1:
                    tmpCompo.addBond(component[1])
                if not tmp.contains(component[0][0]):
                    tmp.addComponent(tmpCompo)
            if not self.contains(tag):
                self.molecules.append(tmp)

    def extend(self, species, update=True):
        if len(self.molecules) == len(species.molecules):
            list1 = sorted(self.molecules, key=lambda x: len(x.components))
            list1 = sorted(list1, key=lambda x: x.name)
            list2 = sorted(species.molecules, key=lambda x: len(x.components))
            list2 = sorted(list2, key=lambda x: x.name)

            for (selement, oelement) in zip(list1, list2):

                cocomponents = Counter([x.name for x in oelement.components])
                for component in oelement.components:
                    refcomponents = Counter([x.name for x in selement.components])
                    # theres not enough components in the base molecule of type 'component'
                    if refcomponents[component.name] < cocomponents[component.name]:
                        selement.components.append(component)
                    else:
                        for element in selement.components:
                            if element.name == component.name:
                                element.addStates(component.states, update)

        else:
            for element in species.molecules:
                if element.name not in [x.name for x in self.molecules]:

                    self.addMolecule(deepcopy(element), update)
                else:

                    bond1 = sum([x.bonds for x in element.components], [])
                    bondList = []
                    for x in [z for z in self.molecules if z.name == element.name]:
                        bondList.append((x, sum([y.bonds for y in x.components], [])))
                    score = 0
                    for x in bondList:
                        if difflib.SequenceMatcher(None, x[1], bond1).ratio() >= score:
                            score = difflib.SequenceMatcher(None, x[1], bond1).ratio()
                            molecule = x[0]
                    # sortedArray = sorted(bondList,
                    #    key=lambda y:difflib.SequenceMatcher(None,y[1],bond1),reverse=True)
                    # molecule = sortedArray[0][0]

                    for component in element.components:
                        if component.name not in [x.name for x in molecule.components]:
                            molecule.addComponent(deepcopy(component), update)
                        else:
                            comp = molecule.getComponent(component.name)
                            for state in component.states:
                                comp.addState(state, update)

    def updateBonds(self, bondNumbers):
        newBondNumbers = deepcopy(bondNumbers)
        correspondence = {}
        intersection = [int(x) for x in newBondNumbers if x in self.getBondNumbers()]
        newBase = max(bondNumbers) + 1
        for element in self.molecules:
            for component in element.components:
                component.bonds = [int(x) + newBase for x in component.bonds]
                """
                for index in range(0,len(component.bonds)):
                    if int(component.bonds[index]) in intersection:

                        if component.bonds[index] in correspondence:
                            component.bonds[index] = correspondence[component.bonds[index]]
                        else:
                            correspondence[component.bonds[index]] = max(intersection) + 1
                            component.bonds[index] = max(intersection) + 1
                """
                # intersection = [int(x) for x in newBondNumbers if x in self.getBondNumbers()]

    def deleteBond(self, moleculePair):
        for molecule in self.molecules:
            if molecule.name in moleculePair:
                moleculePairCopy = deepcopy(moleculePair)
                moleculePairCopy.remove(molecule.name)
                for component in molecule.components:
                    if component.name in [x.lower() for x in moleculePairCopy]:
                        component.bonds = []

    def append(self, species):
        newSpecies = deepcopy(species)
        newSpecies.updateBonds(self.getBondNumbers())

        for element in newSpecies.molecules:
            self.molecules.append(deepcopy(element))

    def sort(self):
        """
        Sort molecules by number of components, then number of bonded components, then lowest numbered bond,
        then the negative sum of the bond index, then number of active states, then string length
        """

        self.molecules.sort(
            key=lambda molecule: (
                len(molecule.components),
                -min(
                    [
                        int(y) if y not in ["?", "+"] else 999
                        for x in molecule.components
                        for y in x.bonds
                    ]
                    + [999]
                ),
                -len([x for x in molecule.components if len(x.bonds) > 0]),
                -len([x for x in molecule.components if x.activeState not in [0, "0"]]),
                len(str(molecule)),
                str(molecule),
            ),
            reverse=True,
        )
        # self.molecules.sort(key=lambda x:(-len(x.components),x.evaluateMolecule(),x.name))

    def __str__(self):
        self.sort()
        return ".".join([x.toString().replace("-", "_") for x in self.molecules])

    def str2(self):
        self.sort()
        return ".".join([x.str2().replace("-", "_") for x in self.molecules])

    def reset(self):
        for element in self.molecules:
            element.reset()

    def toString(self):
        return self.__str__()


import pickle


class Molecule:
    def __init__(self, name):
        """
        Initializes an empty molecule

        >>> str(Molecule('mol'))
        'mol()'

        """
        self.components = []
        self.name = name
        self.compartment = ""
        self.trueName = ""

        a = numpy.random.rand(10, 100)
        self.hash = hashlib.sha1(a).digest()

    def copy(self):
        molecule = Molecule(self.name)
        for element in self.components:
            molecule.components.append(element.copy())
        return molecule

    def addChunk(self, chunk):
        component = Component(chunk[0][0][0][0])
        component.addState(chunk[0][0][0][1])
        self.addComponent(component)

    def addComponent(self, component, overlap=0):
        """
        Adds a component, optionally refusing to add a component if it already exists and sorting components by name
        ---
        Args:
            component (Component): component structure
            overlap (bool): check for overlap before adding

            >>> mol = Molecule('mol')
            >>> mol.addComponent(Component('gamma'))
            >>> mol.addComponent(Component('beta'))
            >>> mol.addComponent(Component('alpha'))
            >>> mol.addComponent(Component('alpha'), overlap=1)
            >>> str(mol)
            'mol(alpha,beta,gamma)'
            >>> mol.addComponent(Component('alpha'), overlap=0)
            >>> str(mol)
            'mol(alpha,alpha,beta,gamma)'
        """
        if not overlap:
            self.components.append(component)
        else:
            if not component.name in [x.name for x in self.components]:
                self.components.append(component)
            else:
                compo = self.getComponent(component.name)
                for state in component.states:
                    compo.addState(state)
        self.components = sorted(self.components, key=lambda x: x.name)

    def setCompartment(self, compartment):
        self.compartment = compartment

    def getBondNumbers(self):
        bondNumbers = []
        for element in self.components:
            bondNumbers.extend([int(x) for x in element.bonds])
        return bondNumbers

    def getComponent(self, componentName):
        for component in self.components:
            if componentName == component.getName():
                return component

    def removeComponent(self, componentName):
        x = [x for x in self.components if x.name == componentName]
        if x != []:
            self.components.remove(x[0])

    def removeComponents(self, components):
        for element in components:
            if element in self.components:
                self.components.remove(element)

    def addBond(self, componentName, bondName):
        bondNumbers = self.getBondNumbers()
        while bondName in bondNumbers:
            bondName += 1
        component = self.getComponent(componentName)
        component.addBond(bondName)

    def getComponentWithBonds(self):
        return [x.name for x in self.components if x.bonds != []]

    def contains(self, componentName):
        return componentName in [x.name for x in self.components]

    def evaluateMolecule(self):
        try:
            return min([self.evaluateBonds(x.bonds) for x in self.components])
        except ValueError:
            return 999999

    def evaluateBonds(self, bonds):
        if len(bonds) == 0:
            return 9999999
        return bonds[0]

    def sort(self):
        self.components.sort(key=lambda x: (x.name, self.evaluateBonds(x.bonds)))

    def __str__(self):
        self.sort()
        tmp = self.name.replace("-", "_")
        if tmp == "0":
            return tmp
        if tmp[0].isdigit():
            tmp = "m__" + tmp
        return (
            tmp
            + "("
            + ",".join([str(x) for x in self.components])
            + ")"
            + self.compartment
        )

    def signature(self, component):
        # tmpComponents = deepcopy(self.components)
        # tmpComponents.sort(key=lambda x:x.name)
        tmp = self.name.replace("-", "_")
        return (
            tmp
            + "("
            + ",".join(
                [
                    x.signature()
                    for x in self.components
                    if (self.name, x.name) not in component
                ]
            )
            + ")"
            + self.compartment
        )

    def toString(self):
        return self.__str__()

    def str2(self):
        self.sort()
        self.components = sorted(self.components, key=lambda x: x.name)
        tmp = self.name.replace("-", "_")
        if tmp[0].isdigit():
            tmp = "m__" + tmp

        return tmp + "(" + ",".join([x.str2() for x in self.components]) + ")"

    def extend(self, molecule):
        for element in molecule.components:
            comp = [x for x in self.components if x.name == element.name]
            if len(comp) == 0:
                self.components.append(deepcopy(element))
            else:
                for bond in element.bonds:
                    comp[0].addBond(bond)
                for state in element.states:
                    comp[0].addState(state)

    def reset(self):
        for element in self.components:
            element.reset()

    def update(self, molecule):
        for comp in molecule.components:
            if comp.name not in [x.name for x in self.components]:
                self.components.append(deepcopy(comp))


class Component:
    def __init__(self, name, bonds=[], states=[]):
        """
        Component initialization setting only the name field
        ---
        Args:
            name (str): The parameter name

            >>> [str(Component('test'))]
            ['test']
            >>> [str(Component('1test'))]
            ['_1test']
        """
        self.name = name
        self.states = []
        self.bonds = []
        self.activeState = ""

    def copy(self):
        """
        Returns an indepent copy of this component

        >>> c = Component('first')
        >>> c2 = c.copy()
        >>> c2.name = 'second'
        >>> [str(c), str(c2)]
        ['first', 'second']
        """
        component = Component(self.name, deepcopy(self.bonds), deepcopy(self.states))
        component.activeState = deepcopy(self.activeState)
        return component

    def addState(self, state, update=True):
        """
        Adds an state to this component, optionally become the active state. The method adds base state 0 if the state
        doesnt exist
        ---
        Args:
            state (str): A string representing the new state variable
            update (bool): A flag indicating whether it should become the new active state (default True)

            >>> c =Component('comp')
            >>> c.addState('first')
            >>> [str(c)]
            ['comp~first']
            >>> c.addState('second', False)
            >>> c.states.sort()
            >>> [str(c), c.str2()]
            ['comp~first', 'comp~0~first~second']
        """
        if not state in self.states:
            self.states.append(state)
        if update:
            self.setActiveState(state)
        if not "0" in self.states:
            self.states.append("0")
        # print 'LALALA',state

    def addStates(self, states, update=True):
        """
        adds a state list to this component, the active state will be the last component in the lsit
        ----
        Args:
            states (list): list of states to add to this component

            >>> c = Component('comp')
            >>> c.addStates(['z1','z2'])
            >>> c.states.sort()
            >>> [str(c), c.str2()]
            ['comp~z2', 'comp~0~z1~z2']
        """
        for state in states:
            if state not in self.states:
                self.addState(state, update)

    def addBond(self, bondName):
        """
        Adds bond information to this component while checking that the bond doesn't already exist
        ---
        Args:
            bondName: Bond identifier to be associated to this component. Its partner is handled at the species level
            >>> c = Component('comp')
            >>> result = c.addBond('1')
            >>> [result, str(c)]
            [True, 'comp!1']
            >>> result = c.addBond('1')
            >>> [result, str(c)]
            [False, 'comp!1']
        """
        # if len(self.bonds) == 0:
        #    self.bonds.append('U')
        if not bondName in self.bonds:
            self.bonds.append(bondName)
            return True
        return False

    def setActiveState(self, state):
        """
        Sets the active state of the component if the state already exists
        ---
        Args:
            state: Sets the new active state

            >>> c = Component('comp')
            >>> c.addState('z1')
            >>> result = c.setActiveState('0')
            >>> [result, str(c)]
            [True, 'comp~0']
            >>> result = c.setActiveState('nonexistant')
            >>> [result, str(c)]
            [False, 'comp~0']
        """
        if state not in self.states:
            return False
        self.activeState = state
        return True

    def getRuleStr(self):
        tmp = self.name
        if len(self.bonds) > 0:
            tmp += "!" + "!".join([str(x) for x in self.bonds])
        if self.activeState != "":
            tmp += "~" + self.activeState
        return tmp

    def getTotalStr(self):
        return self.name + "~".join(self.states)

    def getName(self):
        return self.name

    def __str__(self):
        tmp = self.getRuleStr()
        # edit to meet bng string requirements
        if tmp[0].isdigit():
            tmp = "_" + tmp
        tmp = tmp.replace("-", "_")
        return tmp

    def str2(self):
        tmp = self.name

        if len(self.bonds) > 0:
            tmp += "!" + "!".join([str(x) for x in self.bonds])
        if len(self.states) > 0:
            tmp += "~" + "~".join([str(x) for x in self.states])

        if tmp[0].isdigit():
            tmp = "_" + tmp

        return tmp

    def signature(self):
        tmp = self.name
        if len(self.bonds) > 0:
            tmp += "!+"
        if self.activeState != "":
            tmp += "~{0}".format(self.activeState)
        return tmp

    def __hash__(self):
        return self.name

    def reset(self):
        self.bonds = []
        if "0" in self.states:
            self.activeState = "0"


class Databases:
    def __init__(self):
        # dictionary contains molecule vs bionetgen structure defitions
        self.translator = {}
        self.synthesisDatabase = {}
        self.catalysisDatabase = {}
        self.rawDatabase = {}
        self.labelDictionary = {}
        self.synthesisDatabase2 = {}
        self.assumptions = defaultdict(set)
        self.softConstraints = False
        self.constructedSpecies = set([])
        self.processName = ""

    def getRawDatabase(self):
        return self.rawDatabase

    def getLabelDictionary(self):
        return self.labelDictionary

    def add2LabelDictionary(self, key, value):
        temp = tuple(key)
        temp = temp.sort()
        self.labelDictionary[temp] = value

    def add2RawDatabase(self, rawDatabase):
        pass

    def getTranslator(self):
        return self.translator



import readBNGXML
import argparse
from collections import defaultdict
from rulifier import stateTransitionDiagram


def defineConsole():
    """
    defines the program console line commands
    """
    parser = argparse.ArgumentParser(description="SBML to BNGL translator")
    parser.add_argument("-f1", "--filename1", type=str, help="base file", required=True)
    parser.add_argument(
        "-f2", "--filename2", type=str, help="comparison file", required=True
    )
    return parser


def evaluateStructureSimilarity(moleculeSet1, moleculeSet2):
    scoreDict = defaultdict(list)
    scoreDict["moleculeSetSize"] = len(moleculeSet1)
    for molecule in moleculeSet1:
        for molecule2 in moleculeSet2:
            if molecule.name == molecule2.name:
                scoreDict["overlap"].append(molecule.name)
                # for now we are considering atomizer models so dont check components other than states
                scoreDict["componentSetSize"].extend(
                    [x.name for x in molecule.components if len(x.states) > 0]
                )
                for component in [x for x in molecule.components if len(x.states) > 0]:
                    for component2 in [
                        x for x in molecule2.components if len(x.states) > 0
                    ]:
                        if component.name == component2.name:
                            scoreDict["componentOverlap"].append(component.name)
    return len(scoreDict["overlap"]), scoreDict["moleculeSetSize"]


def evaluateProcessSimilarity(bnglNamespace, bnglNamespace2):
    stdDictionary1 = stateTransitionDiagram.getContextRequirementsFromNamespace(
        bnglNamespace
    )
    stdDictionary2 = stateTransitionDiagram.getContextRequirementsFromNamespace(
        bnglNamespace2
    )
    intersectingNamespace = [
        x.name
        for x in bnglNamespace["molecules"]
        if x.name in [y.name for y in bnglNamespace2["molecules"]]
    ]

    lowerIntersectionNamespace = [x.lower() for x in intersectingNamespace]
    # extract state coverage
    intersectionSpace = defaultdict(set)
    intersectionSpace2 = defaultdict(set)

    scoreDict = {x: {} for x in intersectingNamespace}
    for element in [x for x in stdDictionary1[0] if x in intersectingNamespace]:
        for state in stdDictionary1[0][element]:
            intersectionSpace[element].add(
                frozenset(x for x in state if x[0] in lowerIntersectionNamespace)
            )

    for element in intersectionSpace:
        scoreDict[element]["file1"] = len(intersectionSpace[element])
        scoreDict[element]["totalSpace"] = 2 ** len(
            next(iter(intersectionSpace[element]))
        )
        print element, scoreDict[element]["totalSpace"], next(
            iter(intersectionSpace[element])
        )

    for element in [x for x in stdDictionary2[0] if x in intersectingNamespace]:
        for state in stdDictionary2[0][element]:
            intersectionSpace2[element].add(
                frozenset(x for x in state if x[0] in lowerIntersectionNamespace)
            )

    for element in intersectionSpace:
        scoreDict[element]["file2"] = len(intersectionSpace2[element])

    for element in intersectionSpace:
        scoreDict[element]["intersection"] = [
            len(intersectionSpace[element] - intersectionSpace2[element]),
            len(intersectionSpace2[element] - intersectionSpace[element]),
        ]
        scoreDict[element]["score"] = (
            1
            - scoreDict[element]["intersection"][0] * 1.0 / scoreDict[element]["file1"]
        )
        scoreDict[element]["score2"] = (
            1
            - scoreDict[element]["intersection"][1] * 1.0 / scoreDict[element]["file2"]
        )

    return scoreDict


def evaluateSimilarity(bnglNamespace, bnglNamespace2):
    similarity = {}
    similarity["structure"] = evaluateStructureSimilarity(
        bnglNamespace["molecules"], bnglNamespace2["molecules"]
    )
    similarity["process"] = evaluateProcessSimilarity(bnglNamespace, bnglNamespace2)
    return similarity


if __name__ == "__main__":
    parser = defineConsole()
    namespace = parser.parse_args()

    bnglNamespace = readBNGXML.parseFullXML(namespace.filename1)
    bnglNamespace2 = readBNGXML.parseFullXML(namespace.filename2)
    evaluateSimilarity(bnglNamespace, bnglNamespace2)
